import uuid
import json
import requests
from flask import current_app as app
from flask import session, url_for
from oauthlib.oauth2 import WebApplicationClient

import logging
LOGGER = logging.getLogger(__name__)

def get_provider_cfg():
    return requests.get(app.config["GOOGLE_DISCOVERY_URL"]).json()


def get_client():
    return WebApplicationClient(app.config["GOOGLE_CLIENT_ID"])


def build_auth_url(scopes=None, state=None):
    # Find out what URL to hit for Google login
    google_provider_cfg = get_provider_cfg()
    authorization_endpoint = google_provider_cfg["authorization_endpoint"]
    redirect_uri = url_for("gcallback", _external=True)
    # Use library to construct the request for Google login and provide
    # scopes that let you retrieve user's profile from Google
    return get_client().prepare_request_uri(
        authorization_endpoint,
        redirect_uri=redirect_uri,
        scope=scopes or [],
        state=state or str(uuid.uuid4()))


def get_token(request, code, client):
    client = get_client()
    # Find out what URL to hit to get tokens that allow you to ask for
    # things on behalf of a user
    google_provider_cfg = get_provider_cfg()
    token_endpoint = google_provider_cfg["token_endpoint"]
    # Prepare and send a request to get tokens! Yay tokens!
    token_url, headers, body = client.prepare_token_request(
        token_endpoint,
        authorization_response=request.url,
        redirect_url=request.base_url,
        code=code)

    token_response = requests.post(
        token_url,
        headers=headers,
        data=body,
        auth=(app.config["GOOGLE_CLIENT_ID"], app.config["GOOGLE_CLIENT_SECRET"]))

    # Parse the tokens!
    client.parse_request_body_response(json.dumps(token_response.json()))
    return client


def query_user_info(client):
    # Now that you have tokens (yay) let's find and hit the URL
    # from Google that gives you the user's profile information,
    # including their Google profile image and email
    userinfo_endpoint = get_provider_cfg()["userinfo_endpoint"]
    uri, headers, body = client.add_token(userinfo_endpoint)
    userinfo_response = requests.get(uri, headers=headers, data=body)
    return userinfo_response.json()


def get_token_from_cache():
    token_cache = session.get("token_cache")
    if token_cache is not None:
        return token_cache.get("access_token")



import os
import distutils.util
from inbotauth.user import put_user


class BaseConfig:
    # Application Settings
    SECRET_KEY = os.environ.get('FLASK_OIDC_SECRET_KEY', 'base-dap-config-secret-key')
    WHITELISTED_ENDPOINTS = os.environ.get('FLASK_OIDC_WHITELISTED_ENDPOINTS',
                                           "status,healthcheck,health,login,mslogin,mscallback,"
                                           "glogin,gcallback,static")
    TENANT_BASED_LOGIN = bool(distutils.util.strtobool(os.environ.get('TENANT_BASED_LOGIN', 'True')))
    AUTH_MODELS_MODULE = os.environ.get('AUTH_MODELS_MODULE')
    AUTH_TENANT_CLASS_FUNC = os.environ.get('AUTH_TENANT_CLASS_FUNC')
    AUTH_USER_CLASS_FUNC = os.environ.get('AUTH_USER_CLASS_FUNC')

    # Logging Settings
    LOG_FORMAT = '%(asctime)s.%(msecs)03d [%(levelname)s] %(module)s.%(funcName)s:%(lineno)d (%(process)d:' \
                 + '%(threadName)s) - %(message)s'
    LOG_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S%z'
    LOG_LEVEL = 'INFO'

    OIDC_INTROSPECTION_AUTH_METHOD = 'client_secret_post'
    OIDC_ID_TOKEN_COOKIE_SECURE = False

    # Database and Sessions Settings
    SESSION_TYPE = 'sqlalchemy'
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    SQLALCHEMY_DATABASE_URI = os.environ.get("FLASK_OIDC_SQLALCHEMY_DATABASE_URI", 'sqlite:///sessions.db')
    SQLALCHEMY_ENGINE_OPTIONS = {
        "pool_pre_ping": True
    }

    # Microsoft login
    MS_CLIENT_ID = os.environ.get('MS_CLIENT_ID')
    MS_CLIENT_SECRET = os.environ.get('MS_CLIENT_SECRET')
    MS_USER_ENDPOINT = os.environ.get('MS_USER_ENDPOINT')

    MS_REDIRECT_PATH = "/mscallback"  # Used for forming an absolute URL to your redirect URI.
    # The absolute URL must match the redirect URI you set
    # in the app's registration in the Azure portal.

    # You can find the proper permission names from this document
    # https://docs.microsoft.com/en-us/graph/permissions-reference
    MS_SCOPE = ["User.ReadBasic.All"]

    # Google login
    GOOGLE_CLIENT_ID = os.environ.get('GOOGLE_CLIENT_ID')
    GOOGLE_CLIENT_SECRET = os.environ.get('GOOGLE_CLIENT_SECRET')
    GOOGLE_DISCOVERY_URL = os.environ.get('GOOGLE_DISCOVERY_URL')

    GOOGLE_REDIRECT_PATH = "/gcallback"
    GOOGLE_SCOPE = ["openid", "email", "profile"]

    PUT_USER_METHOD = put_user



import uuid
import requests
from flask import session, url_for
from flask import current_app as app
import msal
from typing import Dict


def load_cache():
    cache = msal.SerializableTokenCache()
    if session.get("token_cache"):
        cache.deserialize(session["token_cache"])
    return cache


def save_cache(cache):
    if cache.has_state_changed:
        session["token_cache"] = cache.serialize()


def build_msal_app(cache=None):
    return msal.ConfidentialClientApplication(app.config['MS_CLIENT_ID'],
                                              client_credential=app.config['MS_CLIENT_SECRET'],
                                              token_cache=cache)


def build_auth_url(scopes=None, state=None):
    redirect_uri = url_for("mscallback", _external=True)
    return build_msal_app().get_authorization_request_url(
        scopes or [],
        state=state or str(uuid.uuid4()),
        redirect_uri=redirect_uri)


def get_token_from_cache(scope=None):
    cache = load_cache()  # This web app maintains one cache per session
    if session.get("token_cache"):
        return cache.find("AccessToken")[0]
    cca = build_msal_app(cache=cache)
    accounts = cca.get_accounts()
    if accounts:  # So all account(s) belong to the current signed-in user
        result = cca.acquire_token_silent(scope, account=accounts[0])
        save_cache(cache)
        return result


def query_user_info(token) -> Dict:
    """
    query user info from oidc server with access token
    """
    return requests.get(  # Use token to call downstream service
        app.config['MS_USER_ENDPOINT'],
        headers={'Authorization': 'Bearer ' + token['secret']},
    ).json()


def get_user(token=None):
    if session.get("auth_user"):
        return session["auth_user"]
    else:
        # if the user is not in session cache, query the user info from OIDC server with the token,
        # and update the user info to local db
        user_info = query_user_info(token)
        user = app.config['PUT_USER_METHOD'](app, user_info)
        session["auth_user"] = user
        return user



import uuid
import logging
import importlib
from flask import Flask, redirect, request, render_template, url_for, session, current_app as app
from flask.helpers import get_env, get_debug_flag
from flask_session import Session
from flask_sqlalchemy import SQLAlchemy

from inbotauth.config import BaseConfig
from inbotauth import azure, google

LOGGER = logging.getLogger(__name__)


class FlaskOIDC(Flask):
    def _before_request(self):
        # ToDo: Need to refactor and divide this method in functions.
        # Whitelisted Endpoints i.e., health checks and status url
        LOGGER.info(f"Request Path: {request.path}")
        LOGGER.info(f"Request Endpoint: {request.endpoint}")
        LOGGER.info(f"Whitelisted Endpoint: {BaseConfig.WHITELISTED_ENDPOINTS}")

        if request.path.strip("/") in BaseConfig.WHITELISTED_ENDPOINTS.split(",") or \
                request.endpoint in BaseConfig.WHITELISTED_ENDPOINTS.split(","):
            return

        # If accepting token in the request headers
        token = None
        if 'Authorization' in request.headers and request.headers['Authorization'].startswith('Bearer '):
            token = request.headers['Authorization'].split(None, 1)[1].strip()
        elif 'access_token' in request.form:
            token = request.form['access_token']
        elif 'access_token' in request.args:
            token = request.args['access_token']
        elif session.get('login_type') == 'microsoft':
            token = azure.get_token_from_cache(self.config['MS_SCOPE'])
        elif session.get('login_type') == 'google':
            token = google.get_token_from_cache()
        if not token:
            return redirect(url_for('login'))

    def __init__(self, *args, **kwargs):
        super(FlaskOIDC, self).__init__(*args, **kwargs)

        # Setup Session Database
        _sql_db = SQLAlchemy(self)
        self.config["SESSION_SQLALCHEMY"] = _sql_db

        auth_models = importlib.import_module(self.config["AUTH_MODELS_MODULE"])
        if self.config['TENANT_BASED_LOGIN']:
            tenant_cls_func = getattr(auth_models, self.config["AUTH_TENANT_CLASS_FUNC"])
            tenant_cls = tenant_cls_func(_sql_db)
        else:
            user_cls_func = getattr(auth_models, self.config["AUTH_USER_CLASS_FUNC"])
            user_cls = user_cls_func(_sql_db)

        # Setup Session Store, that will hold the session information
        # in database. OIDC by default keep the sessions in memory
        _session = Session(self)
        _session.app.session_interface.db.create_all()

        # Register the before request function that will make sure each
        # request is authenticated before processing
        self.before_request(self._before_request)

        self.jinja_env.globals.update(_build_auth_url=azure.build_auth_url)  # Used in template

        @self.route('/login')  # catch_all
        def login():
            return render_template("login.html")

        @self.route('/mslogin')  # catch_all
        def mslogin():
            session["state"] = str(uuid.uuid4())
            # Technically we could use empty list [] as scopes to do just sign in,
            # here we choose to also collect end user consent upfront
            auth_url = azure.build_auth_url(scopes=self.config['MS_SCOPE'], state=session["state"])
            return redirect(auth_url)

        @self.route('/mscallback')  # catch_all
        def mscallback():
            code = request.args.get('code')
            if request.args.get('state') != session.get("state"):
                return redirect(url_for('index'))  # No-OP. Goes back to Index page
            if "error" in request.args:  # Authentication/Authorization failure
                return render_template("auth_error.html", result=request.args)
            if code:
                cache = azure.load_cache()
                redirect_uri = url_for("mscallback", _external=True)
                result = azure.build_msal_app(cache=cache).acquire_token_by_authorization_code(
                    code,
                    scopes=self.config['MS_SCOPE'],  # Misspelled scope would cause an HTTP 400 error here
                    redirect_uri=redirect_uri)
                if "error" in result:
                    return render_template("auth_error.html", result=result)
                user_data = result.get("id_token_claims")
                if not user_data:
                    LOGGER.info("failed to get user info for state {}, code {}".format(
                        request.args.get('state'), code))
                    return render_template("auth_error.html", result=result)

                user_info = azure.query_user_info(cache.find("AccessToken")[0])
                domain = user_info['email'].split('@')[1]
                user_info['domain'] = domain
                user_info['tenant_id'] = user_data.get('tid')

                # the user is authenticated only if successfully adding the user
                user = self.config['PUT_USER_METHOD'](self, user_info)
                if not user:
                    return render_template("auth_error.html", result=request.args)

                # authenticate against database
                if self.config['TENANT_BASED_LOGIN']:
                    # if tenant based login, check if user tenant exists in tenant table
                    has_access = self.authenticate(tenant_cls, domain, result)
                else:
                    # if user based login, check if user exists in user table
                    has_access = self.authenticate(user_cls, user['email'], result)
                if not has_access:
                    return render_template("auth_error.html", result=user_info)

                azure.save_cache(cache)
                session["auth_user"] = user
                session['login_type'] = 'microsoft'
            return redirect(url_for('index'))

        @self.route('/glogin')  # catch_all
        def glogin():
            session["state"] = str(uuid.uuid4())
            # Technically we could use empty list [] as scopes to do just sign in,
            # here we choose to also collect end user consent upfront
            auth_url = google.build_auth_url(scopes=self.config['GOOGLE_SCOPE'], state=session["state"])
            return redirect(auth_url)

        @self.route("/gcallback")
        def gcallback():
            code = request.args.get('code')
            if request.args.get('state') != session.get("state"):
                return redirect(url_for('index'))  # No-OP. Goes back to Index page
            if "error" in request.args:  # Authentication/Authorization failure
                return render_template("auth_error.html", result=request.args)
            if code:
                client = google.get_client()
                client = google.get_token(request, code, client)
                user_info = google.query_user_info(client)
                domain = user_info['hd']
                user_info['domain'] = domain

                # the user is authenticated only if successfully adding the user
                user = self.config['PUT_USER_METHOD'](self, user_info)
                if not user:
                    return render_template("auth_error.html", result=request.args)

                # authenticate against database
                if self.config['TENANT_BASED_LOGIN']:
                    # if tenant based login, check if user domain exists in tenant table
                    has_access = self.authenticate(tenant_cls, domain, user_info)
                else:
                    # if user based login, check if user exists in user table
                    has_access = self.authenticate(user_cls, user['email'], user_info)
                if not has_access:
                    return render_template("auth_error.html", result=user_info)

                session["auth_user"] = user
                session['login_type'] = 'google'
                session['token_cache'] = client.token
            return redirect(url_for('index'))

        @self.route('/logout')  # catch_all
        def logout():
            if session['login_type'] == 'microsoft':
                authority = 'https://login.microsoftonline.com/' + session['auth_user']['tenant_id']
                session.clear()  # Wipe out user and its token cache from session
                return redirect(  # Also logout from your tenant's web session
                    authority + "/oauth2/v2.0/logout" + "?post_logout_redirect_uri=" \
                    + url_for("index", _external=True))
            elif session['login_type'] == 'google':
                session.clear()
                return redirect(url_for('login'))

    def authenticate(self, model_class, key, error_text):
        if model_class.query.get(key) is None:
            LOGGER.info("Authentication failed: {}".format(error_text))
            return False
        return True

    def make_config(self, instance_relative=False):
        """
        Overriding the default `make_config` function in order to support
        Flask OIDC package and all of their settings.
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        defaults = dict(self.default_config)
        defaults['ENV'] = get_env()
        defaults['DEBUG'] = get_debug_flag()

        # Append all the configurations from the base config class.
        for key, value in BaseConfig.__dict__.items():
            if not key.startswith('__'):
                defaults[key] = value
        return self.config_class(root_path, defaults)



from setuptools import setup, find_packages

__version__ = '1.2.2'

setup(
    name='inbot-auth',
    version=__version__,
    description='Flask wrapper with pre-configured azure OIDC support',
    url='https://github.com/Inbot/inbotauth.git',
    maintainer='Aarni Alasaarela',
    maintainer_email='aarni.alasaarela@gmail.com',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    dependency_links=[],
    install_requires=[
        'Flask>=1.0.2',
        'Flask-Session>=0.3.0',
        'Flask-SQLAlchemy>=2.4.0',
        'marshmallow>=2.21.0',
        'msal>=1.8.0',
        'requests>=2.24.0',
        'SQLAlchemy>=1.3.22',
        'urllib3>=1.21.1',
        'Werkzeug>=0.14'
    ],
    python_requires=">=3.6"
)



from flask import current_app as app
from flask import session
from inbotauth import azure, google
import logging

LOGGER = logging.getLogger(__name__)


def get_token_from_cache():
    LOGGER.info(session.get('login_type'))
    if session.get('login_type') == 'google':
        return google.get_token_from_cache()
    if session.get('login_type') == 'microsoft':
        return azure.get_token_from_cache(scope=app.config['MS_SCOPE'])['secret']
    return None
